local OnEvent = {};
local FrameAnchorBuffer = { PreMove = {}, PostMove = {}, PreSize = {}, IsDragging = false };

local AddOn_Loaded = false;
local Hooked = false;

local SettingsFrame;
local TalkingHeadDragFrame;
local TalkingHeadDragAnchorFrame;

-- Safe start moving function for anchored frames.
function MoveAnchoredFrameStart(self)
	if(not FrameAnchorBuffer.IsDragging) then
		FrameAnchorBuffer.IsDragging = true;
		FrameAnchorBuffer.PreMove.Point, FrameAnchorBuffer.PreMove.RelativeTo, FrameAnchorBuffer.PreMove.RelativePoint, FrameAnchorBuffer.PreMove.OffsetX, FrameAnchorBuffer.PreMove.OffsetY = self:GetPoint();
		self:StartMoving();
		FrameAnchorBuffer.PostMove.Point, FrameAnchorBuffer.PostMove.RelativeTo, FrameAnchorBuffer.PostMove.RelativePoint, FrameAnchorBuffer.PostMove.OffsetX, FrameAnchorBuffer.PostMove.OffsetY = self:GetPoint();
	end
end

-- Safe stop moving function for anchored frames.
function MoveAnchoredFrameStop(self)
	if(FrameAnchorBuffer.IsDragging) then
		FrameAnchorBuffer.IsDragging = false;
		local point, relative_too, relative_point, offset_x, offset_y = self:GetPoint();
		self:StopMovingOrSizing();
		self:ClearAllPoints();
		self:SetPoint(FrameAnchorBuffer.PreMove.Point, FrameAnchorBuffer.PreMove.RelativeTo, FrameAnchorBuffer.PreMove.RelativePoint, (FrameAnchorBuffer.PreMove.OffsetX + (offset_x - FrameAnchorBuffer.PostMove.OffsetX)), (FrameAnchorBuffer.PreMove.OffsetY + (offset_y - FrameAnchorBuffer.PostMove.OffsetY)));
	end
end

local function THFUpdateScale()
	if(Hooked and AddOn_Loaded) then
		TalkingHeadFrame:SetScale(THF_DB.Scale);
	end
end

local function THFUpdateClickThrough()
	if(Hooked and AddOn_Loaded) then
		if(TalkingHeadDragFrame:IsVisible()) then
			TalkingHeadFrame:EnableMouse(false);
		else
			TalkingHeadFrame:EnableMouse(not THF_DB.ClickThrough);
		end
	end
end

local function THFUpdateVisibility()
	if(Hooked and AddOn_Loaded) then
		if(TalkingHeadFrame:IsVisible() and THF_DB.Disabled) then
			TalkingHeadFrame.MainFrame.CloseButton:GetScript("OnClick")();
		end
	end
end

local function THFUpdatePosition()
	if(Hooked and AddOn_Loaded) then
		TalkingHeadFrame:ClearAllPoints();
		if(TalkingHeadDragFrame:IsVisible()) then
			TalkingHeadFrame:SetPoint("CENTER", TalkingHeadDragFrame, "CENTER", 0, 0);
		else
			TalkingHeadFrame:SetPoint("CENTER", TalkingHeadDragAnchorFrame, "CENTER", 0, 0);
		end
	end
end

local function THFUpdateStrata()
	if(Hooked and AddOn_Loaded) then
		if(THF_DB.LowerStrata and not TalkingHeadDragFrame:IsVisible()) then
			TalkingHeadFrame:SetFrameStrata("BACKGROUND");
		else
			TalkingHeadFrame:SetFrameStrata("HIGH");
		end
	end
end

local function THFOnShow()
	THFUpdateScale();
	THFUpdatePosition();
	THFUpdateClickThrough();
	THFUpdateVisibility();
	THFUpdateStrata();
end

local function HookTHF()
	if not Hooked and TalkingHeadFrame and TalkingHeadFrame:GetScript("OnShow") then
		TalkingHeadFrame:HookScript("OnShow", THFOnShow);
		Hooked = true;
		THFOnShow();
	end
end

OnEvent["TALKINGHEAD_REQUESTED"] = function(self, ...)
	HookTHF();
	THFOnShow();
end

OnEvent["LOADING_SCREEN_DISABLED"] = function(self, ...)
	HookTHF(); -- Not sure if this line is neccessary, but its for good measure.
	THFOnShow();
end

OnEvent["ADDON_LOADED"] = function(self, ...)
	if(not AddOn_Loaded) then
		THF_DB = THF_DB or {};
		
		-- Load or use default values
		THF_DB.X = THF_DB.X or 0;
		THF_DB.Y = THF_DB.Y or 0;
		THF_DB.Scale = THF_DB.Scale or 1;
		THF_DB.Point = THF_DB.Point or "CENTER";
		THF_DB.RelativePoint = THF_DB.RelativePoint or "CENTER";
		THF_DB.ClickThrough = THF_DB.ClickThrough or false;
		THF_DB.Disabled = THF_DB.Disabled or false;
		THF_DB.LowerStrata = THF_DB.LowerStrata or false;
		
		-- Initialize values.
		SettingsFrame.CBEnable:SetChecked(not THF_DB.Disabled);
		SettingsFrame.CBClickThrough:SetChecked(THF_DB.ClickThrough);
		SettingsFrame.CBLowerStrata:SetChecked(THF_DB.LowerStrata);
		
		-- Set the scale for the drag frame & settings frame.
		TalkingHeadDragFrame:SetScale(THF_DB.Scale);
		SettingsFrame:SetScale(1 / THF_DB.Scale);
		SettingsFrame.ScaleSlider:SetValue(THF_DB.Scale * 100);
		getglobal(SettingsFrame.ScaleSlider:GetName() .. 'Text'):SetText(format("|cffffff00Scale: %.0f%%|r", SettingsFrame.ScaleSlider:GetValue()));
		
		-- Set the location of the drag frame.
		TalkingHeadDragAnchorFrame:ClearAllPoints();
		TalkingHeadDragAnchorFrame:SetPoint(THF_DB.Point, UIParent, THF_DB.RelativePoint, THF_DB.X, THF_DB.Y);
		
		-- Set the addon loaded flag.
		AddOn_Loaded = true;
		
		-- Fire the on show function in case the frame is already displayed.
		THFOnShow();
	end
end

local function SetToolTip(frame, title, text)
	if(text or title) then
		frame:HookScript("OnLeave", function(self)
			GameTooltip:Hide();
		end);
		frame:HookScript("OnEnter", function(self)
			GameTooltip:Hide();
			GameTooltip:SetOwner(self, "ANCHOR_CURSOR");
			if(not title) then
				GameTooltip:SetText(text, 1, 1, 1);
			else
				GameTooltip:SetText(title);
				GameTooltip:AddLine(text, 1, 1, 1);
			end
			GameTooltip:Show();
		end);
	end

end

local function Initialize()
	-- Event frame
	local frame = CreateFrame("Frame");
	frame:SetScript("OnEvent", function(self, event, ...) OnEvent[event](self, ...); end);
	for event, func in pairs(OnEvent) do
		frame:RegisterEvent(event);
	end
	
	TalkingHeadDragAnchorFrame = CreateFrame("Frame", "TalkingHeadDragFrame", UIParent);
	TalkingHeadDragAnchorFrame:Show();
	TalkingHeadDragAnchorFrame:SetWidth(1);
	TalkingHeadDragAnchorFrame:SetHeight(1);
	TalkingHeadDragAnchorFrame:EnableMouse(true);
	TalkingHeadDragAnchorFrame:SetMovable(true);
	TalkingHeadDragAnchorFrame:SetUserPlaced(false);
	TalkingHeadDragAnchorFrame:RegisterForDrag("LeftButton");
	TalkingHeadDragAnchorFrame:ClearAllPoints();
	TalkingHeadDragAnchorFrame:SetPoint("CENTER", UIParent, "CENTER", 0, 0);
	
	-- Drag frame
	TalkingHeadDragFrame = CreateFrame("Frame", "TalkingHeadDragFrame", TalkingHeadDragAnchorFrame);
	TalkingHeadDragFrame:Hide();
	TalkingHeadDragFrame:SetBackdrop({ bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background", 
		edgeFile = nil,
		tile = nil,
		tileSize = 0,
		edgeSize = 0,
		insets = { left = 0, right = 0, top = 0, bottom = 0 }});
	TalkingHeadDragFrame:SetFrameStrata("HIGH");
	TalkingHeadDragFrame:EnableMouse(true);
	TalkingHeadDragFrame:SetMovable(true);
	TalkingHeadDragFrame:SetUserPlaced(false);
	TalkingHeadDragFrame:RegisterForDrag("LeftButton");
	TalkingHeadDragFrame.THFText = TalkingHeadDragFrame:CreateFontString();
	TalkingHeadDragFrame.THFText:SetFont("Fonts\\FRIZQT__.TTF", 38, "OUTLINE");
	TalkingHeadDragFrame.THFText:SetPoint("CENTER", 0, 0);
	TalkingHeadDragFrame.THFText:SetText("(TalkingHeadFrame)");
	TalkingHeadDragFrame.MoveText = TalkingHeadDragFrame:CreateFontString();
	TalkingHeadDragFrame.MoveText:SetFont("Fonts\\FRIZQT__.TTF", 14, "OUTLINE");
	TalkingHeadDragFrame.MoveText:SetPoint("CENTER", 0, 40);
	TalkingHeadDragFrame.MoveText:SetText("(Click & Drag To Move)");
	TalkingHeadDragFrame:SetWidth(543); -- 570
	TalkingHeadDragFrame:SetHeight(128); -- 150
	TalkingHeadDragFrame:ClearAllPoints();
	TalkingHeadDragFrame:SetPoint("CENTER", TalkingHeadDragAnchorFrame, "CENTER", 0, 0);
	TalkingHeadDragFrame:SetScript("OnShow", THFOnShow);
	TalkingHeadDragFrame:SetScript("OnHide", THFOnShow);
	TalkingHeadDragFrame:SetScript("OnMouseDown", function(self, ...)
		THF_DB.Point, _, THF_DB.RelativePoint, THF_DB.X, THF_DB.Y = TalkingHeadDragAnchorFrame:GetPoint();
		MoveAnchoredFrameStart(TalkingHeadDragAnchorFrame);
		THFUpdatePosition();
	end);
	TalkingHeadDragFrame:SetScript("OnMouseUp", function(self, ...)
		MoveAnchoredFrameStop(TalkingHeadDragAnchorFrame);
		THF_DB.Point, _, THF_DB.RelativePoint, THF_DB.X, THF_DB.Y = TalkingHeadDragAnchorFrame:GetPoint();
		THFUpdatePosition();
	end);
	
	-- Settings frame.
	SettingsFrame = CreateFrame("Frame", "TalkingHeadDragFrame", TalkingHeadDragFrame);
	SettingsFrame:Show();
	SettingsFrame:SetBackdrop({ bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background", 
			edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
			tile = nil,
			tileSize = 0,
			edgeSize = 14,
			insets = { left = 3, right = 3, top = 3, bottom = 3 }});
	SettingsFrame:SetHeight(137);
	SettingsFrame:SetWidth(250);
	SettingsFrame:SetFrameStrata("HIGH");
	SettingsFrame:EnableMouse(true);
	SettingsFrame:ClearAllPoints();
	SettingsFrame:SetPoint("TOP", TalkingHeadDragFrame, "BOTTOM", 0, -5)
	
	SettingsFrame.Ttitle = SettingsFrame:CreateFontString();
	SettingsFrame.Ttitle:SetFont("Fonts\\FRIZQT__.TTF", 16, "OUTLINE");
	SettingsFrame.Ttitle:SetPoint("TOP", 0, -6);
	SettingsFrame.Ttitle:SetText("|cffffff00TalkingHeadFrame|r |cff00ffffSettings|r");
	
	SettingsFrame.ScaleSlider = CreateFrame("Slider", "THFSettingsFrameScaleSlider", SettingsFrame, "OptionsSliderTemplate");
	--SetToolTip(SettingsFrame.ScaleSlider, "Scale", "Magnify or shrink the TalkingHeadFrame."); -- Blocks the slider at the moment
	SettingsFrame.ScaleSlider:SetWidth(210);
	SettingsFrame.ScaleSlider:SetHeight(16);
	SettingsFrame.ScaleSlider:SetPoint("TOPLEFT", 18, -40);
	SettingsFrame.ScaleSlider:SetOrientation("HORIZONTAL");
	getglobal(SettingsFrame.ScaleSlider:GetName() .. 'Low'):SetText();
	getglobal(SettingsFrame.ScaleSlider:GetName() .. 'High'):SetText();
	getglobal(SettingsFrame.ScaleSlider:GetName() .. 'Text'):SetText("|cffffff00Scale: 100%|r");
	--getglobal(SettingsFrame.ScaleSlider:GetName() .. 'Low'):SetFont("Fonts\\FRIZQT__.TTF", 14, "OUTLINE");
	--getglobal(SettingsFrame.ScaleSlider:GetName() .. 'High'):SetFont("Fonts\\FRIZQT__.TTF", 14, "OUTLINE");
	getglobal(SettingsFrame.ScaleSlider:GetName() .. 'Text'):SetFont("Fonts\\FRIZQT__.TTF", 14, "OUTLINE");
	SettingsFrame.ScaleSlider:SetMinMaxValues(40, 200);
	SettingsFrame.ScaleSlider:SetValue(100);
	SettingsFrame.ScaleSlider:SetScript("OnValueChanged", function()
		getglobal(SettingsFrame.ScaleSlider:GetName() .. 'Text'):SetText(format("|cffffff00Scale: %.0f%%|r", SettingsFrame.ScaleSlider:GetValue()));
		THF_DB.Scale = SettingsFrame.ScaleSlider:GetValue() / 100;
		TalkingHeadDragFrame:SetScale(THF_DB.Scale);
		SettingsFrame:SetScale(1 / THF_DB.Scale);
		THFUpdateScale();
		THFUpdatePosition();
	end);
	
	SettingsFrame.CBEnable = CreateFrame("CheckButton", "THFSettingsFrameCBEnable", SettingsFrame, "UICheckButtonTemplate");
	SettingsFrame.CBEnable:SetWidth(25);
	SettingsFrame.CBEnable:SetHeight(25);
	SetToolTip(SettingsFrame.CBEnable, "Enable", "Untick to have the TalkingHeadFrame never show.");
	getglobal(SettingsFrame.CBEnable:GetName() .. "Text"):SetFont("Fonts\\FRIZQT__.TTF", 14, "OUTLINE");
	getglobal(SettingsFrame.CBEnable:GetName() .. "Text"):SetText("Enabled")
	SettingsFrame.CBEnable:SetPoint("TOPLEFT", 18, -55)
	SettingsFrame.CBEnable:SetScript("PostClick", function()
		THF_DB.Disabled = not SettingsFrame.CBEnable:GetChecked();
		THFUpdateVisibility();
	end);
	
	SettingsFrame.CBClickThrough = CreateFrame("CheckButton", "THFSettingsFrameCBClickThrough", SettingsFrame, "UICheckButtonTemplate");
	SettingsFrame.CBClickThrough:SetWidth(25);
	SettingsFrame.CBClickThrough:SetHeight(25);
	SetToolTip(SettingsFrame.CBClickThrough, "Click Through", "Allows you to interact with stuff behind the TalkingHeadFrame.");
	getglobal(SettingsFrame.CBClickThrough:GetName() .. "Text"):SetFont("Fonts\\FRIZQT__.TTF", 14, "OUTLINE");
	getglobal(SettingsFrame.CBClickThrough:GetName() .. "Text"):SetText("Click Through")
	SettingsFrame.CBClickThrough:SetPoint("TOPLEFT", 18, -80)
	SettingsFrame.CBClickThrough:SetScript("PostClick", function()
		THF_DB.ClickThrough = SettingsFrame.CBClickThrough:GetChecked();
		THFUpdateClickThrough();
	end);
	
	SettingsFrame.CBLowerStrata = CreateFrame("CheckButton", "THFSettingsFrameCBLowerStrata", SettingsFrame, "UICheckButtonTemplate");
	SettingsFrame.CBLowerStrata:SetWidth(25);
	SettingsFrame.CBLowerStrata:SetHeight(25);
	SetToolTip(SettingsFrame.CBLowerStrata, "Lower Frame Strata", "TalkingHeadFrame will now appear behind the majority of other AddOns.");
	getglobal(SettingsFrame.CBLowerStrata:GetName() .. "Text"):SetFont("Fonts\\FRIZQT__.TTF", 14, "OUTLINE");
	getglobal(SettingsFrame.CBLowerStrata:GetName() .. "Text"):SetText("Lower Frame Strata")
	SettingsFrame.CBLowerStrata:SetPoint("TOPLEFT", 18, -105)
	SettingsFrame.CBLowerStrata:SetScript("PostClick", function()
		THF_DB.LowerStrata = SettingsFrame.CBLowerStrata:GetChecked();
		THFUpdateStrata();
	end);
	
	-- Slash command handler.
	SLASH_TALKINGHEADFRAME1 = "/thf";
	SLASH_TALKINGHEADFRAME2 = "/talkingheadframe";
	SlashCmdList["TALKINGHEADFRAME"] = function(msg, editbox)
		local lower_msg = msg:lower();
		
		if(lower_msg == "reset") then
			THF_DB.X = 0;
			THF_DB.Y = 0;
			THF_DB.Point = "CENTER";
			THF_DB.RelativePoint = "CENTER";
			TalkingHeadDragAnchorFrame:ClearAllPoints();
			TalkingHeadDragAnchorFrame:SetPoint(THF_DB.Point, UIParent, THF_DB.RelativePoint, THF_DB.X, THF_DB.Y);
			THFOnShow();
		else
			if(TalkingHeadDragFrame:IsVisible()) then
				TalkingHeadDragFrame:Hide();
			else
				TalkingHeadDragFrame:Show();
			end
		end
	end
end

Initialize();